@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2012-2013 250bpm s.r.o.  All rights reserved.
+    Copyright (c) 2012-2013 Martin Sustrik  All rights reserved.
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"),
@@ -63,60 +63,6 @@
 
 #define NN_REQ_SRC_RESEND_TIMER 1
 
-struct nn_req {
-
-    /*  The base class. Raw REQ socket. */
-    struct nn_xreq xreq;
-
-    /*  The state machine. */
-    struct nn_fsm fsm;
-    int state;
-
-    /*  ID of the request being currently processed. Replies for different
-        requests are considered stale and simply dropped. */
-    uint32_t reqid;
-
-    /*  Stored request, so that it can be re-sent if needed. */
-    struct nn_msg request;
-
-    /*  Stored reply, so that user can retrieve it later on. */
-    struct nn_msg reply;
-
-    /*  Timer used to wait while request should be re-sent. */
-    struct nn_timer timer;
-
-    /*  Protocol-specific socket options. */
-    int resend_ivl;
-
-    /*  Pipe the current request has been sent to. Non-null only in ACTIVE
-        state  */
-    struct nn_pipe *sent_to;
-};
-
-/*  Private functions. */
-static void nn_req_init (struct nn_req *self,
-    const struct nn_sockbase_vfptr *vfptr, void *hint);
-static void nn_req_term (struct nn_req *self);
-static int nn_req_inprogress (struct nn_req *self);
-static void nn_req_handler (struct nn_fsm *self, int src, int type,
-    void *srcptr);
-static void nn_req_shutdown (struct nn_fsm *self, int src, int type,
-    void *srcptr);
-static void nn_req_action_send (struct nn_req *self, int allow_delay);
-
-/*  Implementation of nn_sockbase's virtual functions. */
-static void nn_req_stop (struct nn_sockbase *self);
-static void nn_req_destroy (struct nn_sockbase *self);
-static void nn_req_in (struct nn_sockbase *self, struct nn_pipe *pipe);
-static void nn_req_out (struct nn_sockbase *self, struct nn_pipe *pipe);
-static int nn_req_events (struct nn_sockbase *self);
-static int nn_req_send (struct nn_sockbase *self, struct nn_msg *msg);
-static void nn_req_rm (struct nn_sockbase *self, struct nn_pipe *pipe);
-static int nn_req_recv (struct nn_sockbase *self, struct nn_msg *msg);
-static int nn_req_setopt (struct nn_sockbase *self, int level, int option,
-    const void *optval, size_t optvallen);
-static int nn_req_getopt (struct nn_sockbase *self, int level, int option,
-    void *optval, size_t *optvallen);
 static const struct nn_sockbase_vfptr nn_req_sockbase_vfptr = {
     nn_req_stop,
     nn_req_destroy,
@@ -125,44 +71,52 @@ static const struct nn_sockbase_vfptr nn_req_sockbase_vfptr = {
     nn_req_in,
     nn_req_out,
     nn_req_events,
-    nn_req_send,
-    nn_req_recv,
+    nn_req_csend,
+    nn_req_crecv,
     nn_req_setopt,
     nn_req_getopt
 };
 
-static void nn_req_init (struct nn_req *self,
+void nn_req_init (struct nn_req *self,
     const struct nn_sockbase_vfptr *vfptr, void *hint)
 {
+    nn_req_handle hndl;
+
     nn_xreq_init (&self->xreq, vfptr, hint);
     nn_fsm_init_root (&self->fsm, nn_req_handler, nn_req_shutdown,
         nn_sockbase_getctx (&self->xreq.sockbase));
     self->state = NN_REQ_STATE_IDLE;
-    self->sent_to = NULL;
 
     /*  Start assigning request IDs beginning with a random number. This way
         there should be no key clashes even if the executable is re-started. */
-    nn_random_generate (&self->reqid, sizeof (self->reqid));
+    nn_random_generate (&self->lastid, sizeof (self->lastid));
 
-    nn_msg_init (&self->request, 0);
-    nn_msg_init (&self->reply, 0);
-    nn_timer_init (&self->timer, NN_REQ_SRC_RESEND_TIMER, &self->fsm);
+    self->task.sent_to = NULL;
+
+    nn_msg_init (&self->task.request, 0);
+    nn_msg_init (&self->task.reply, 0);
+    nn_timer_init (&self->task.timer, NN_REQ_SRC_RESEND_TIMER, &self->fsm);
     self->resend_ivl = NN_REQ_DEFAULT_RESEND_IVL;
 
+    /*  For now, handle is empty. */
+    memset (&hndl, 0, sizeof (hndl));
+    nn_task_init (&self->task, self->lastid, hndl);
+
     /*  Start the state machine. */
     nn_fsm_start (&self->fsm);
 }
 
-static void nn_req_term (struct nn_req *self)
+void nn_req_term (struct nn_req *self)
 {
-    nn_timer_term (&self->timer);
-    nn_msg_term (&self->reply);
-    nn_msg_term (&self->request);
+    nn_timer_term (&self->task.timer);
+    nn_task_term (&self->task);
+    nn_msg_term (&self->task.reply);
+    nn_msg_term (&self->task.request);
     nn_fsm_term (&self->fsm);
     nn_xreq_term (&self->xreq);
 }
 
-static void nn_req_stop (struct nn_sockbase *self)
+void nn_req_stop (struct nn_sockbase *self)
 {
     struct nn_req *req;
 
@@ -171,7 +125,7 @@ static void nn_req_stop (struct nn_sockbase *self)
     nn_fsm_stop (&req->fsm);
 }
 
-static void nn_req_destroy (struct nn_sockbase *self)
+void nn_req_destroy (struct nn_sockbase *self)
 {
     struct nn_req *req;
 
@@ -181,7 +135,7 @@ static void nn_req_destroy (struct nn_sockbase *self)
     nn_free (req);
 }
 
-static int nn_req_inprogress (struct nn_req *self)
+int nn_req_inprogress (struct nn_req *self)
 {
     /*  Return 1 if there's a request submitted. 0 otherwise. */
     return self->state == NN_REQ_STATE_IDLE ||
@@ -189,7 +143,7 @@ static int nn_req_inprogress (struct nn_req *self)
         self->state == NN_REQ_STATE_STOPPING ? 0 : 1;
 }
 
-static void nn_req_in (struct nn_sockbase *self, struct nn_pipe *pipe)
+void nn_req_in (struct nn_sockbase *self, struct nn_pipe *pipe)
 {
     int rc;
     struct nn_req *req;
@@ -203,37 +157,38 @@ static void nn_req_in (struct nn_sockbase *self, struct nn_pipe *pipe)
     while (1) {
 
         /*  Get new reply. */
-        rc = nn_xreq_recv (&req->xreq.sockbase, &req->reply);
+        rc = nn_xreq_recv (&req->xreq.sockbase, &req->task.reply);
         if (nn_slow (rc == -EAGAIN))
             return;
         errnum_assert (rc == 0, -rc);
 
         /*  No request was sent. Getting a reply doesn't make sense. */
         if (nn_slow (!nn_req_inprogress (req))) {
-            nn_msg_term (&req->reply);
+            nn_msg_term (&req->task.reply);
             continue;
         }
 
         /*  Ignore malformed replies. */
-        if (nn_slow (nn_chunkref_size (&req->reply.hdr) != sizeof (uint32_t))) {
-            nn_msg_term (&req->reply);
+        if (nn_slow (nn_chunkref_size (&req->task.reply.sphdr) !=
+              sizeof (uint32_t))) {
+            nn_msg_term (&req->task.reply);
             continue;
         }
 
         /*  Ignore replies with incorrect request IDs. */
-        reqid = nn_getl (nn_chunkref_data (&req->reply.hdr));
+        reqid = nn_getl (nn_chunkref_data (&req->task.reply.sphdr));
         if (nn_slow (!(reqid & 0x80000000))) {
-            nn_msg_term (&req->reply);
+            nn_msg_term (&req->task.reply);
             continue;
         }
-        if (nn_slow (reqid != (req->reqid | 0x80000000))) {
-            nn_msg_term (&req->reply);
+        if (nn_slow (reqid != (req->task.id | 0x80000000))) {
+            nn_msg_term (&req->task.reply);
             continue;
         }
 
         /*  Trim the request ID. */
-        nn_chunkref_term (&req->reply.hdr);
-        nn_chunkref_init (&req->reply.hdr, 0);
+        nn_chunkref_term (&req->task.reply.sphdr);
+        nn_chunkref_init (&req->task.reply.sphdr, 0);
 
         /*  TODO: Deallocate the request here? */
 
@@ -245,7 +200,7 @@ static void nn_req_in (struct nn_sockbase *self, struct nn_pipe *pipe)
     }
 }
 
-static void nn_req_out (struct nn_sockbase *self, struct nn_pipe *pipe)
+void nn_req_out (struct nn_sockbase *self, struct nn_pipe *pipe)
 {
     struct nn_req *req;
 
@@ -259,7 +214,7 @@ static void nn_req_out (struct nn_sockbase *self, struct nn_pipe *pipe)
         nn_fsm_action (&req->fsm, NN_REQ_ACTION_OUT);
 }
 
-static int nn_req_events (struct nn_sockbase *self)
+int nn_req_events (struct nn_sockbase *self)
 {
     int rc;
     struct nn_req *req;
@@ -277,7 +232,13 @@ static int nn_req_events (struct nn_sockbase *self)
     return rc;
 }
 
-static int nn_req_send (struct nn_sockbase *self, struct nn_msg *msg)
+int nn_req_send (NN_UNUSED int s, NN_UNUSED nn_req_handle hndl,
+    NN_UNUSED const void *buf, NN_UNUSED size_t len, NN_UNUSED int flags)
+{
+    nn_assert (0);
+}
+
+int nn_req_csend (struct nn_sockbase *self, struct nn_msg *msg)
 {
     struct nn_req *req;
 
@@ -286,15 +247,15 @@ static int nn_req_send (struct nn_sockbase *self, struct nn_msg *msg)
     /*  Generate new request ID for the new request and put it into message
         header. The most important bit is set to 1 to indicate that this is
         the bottom of the backtrace stack. */
-    ++req->reqid;
-    nn_assert (nn_chunkref_size (&msg->hdr) == 0);
-    nn_chunkref_term (&msg->hdr);
-    nn_chunkref_init (&msg->hdr, 4);
-    nn_putl (nn_chunkref_data (&msg->hdr), req->reqid | 0x80000000);
+    ++req->task.id;
+    nn_assert (nn_chunkref_size (&msg->sphdr) == 0);
+    nn_chunkref_term (&msg->sphdr);
+    nn_chunkref_init (&msg->sphdr, 4);
+    nn_putl (nn_chunkref_data (&msg->sphdr), req->task.id | 0x80000000);
 
     /*  Store the message so that it can be re-sent if there's no reply. */
-    nn_msg_term (&req->request);
-    nn_msg_mv (&req->request, msg);
+    nn_msg_term (&req->task.request);
+    nn_msg_mv (&req->task.request, msg);
 
     /*  Notify the state machine. */
     nn_fsm_action (&req->fsm, NN_REQ_ACTION_SENT);
@@ -302,7 +263,14 @@ static int nn_req_send (struct nn_sockbase *self, struct nn_msg *msg)
     return 0;
 }
 
-static int nn_req_recv (struct nn_sockbase *self, struct nn_msg *msg)
+int nn_req_recv (NN_UNUSED int s, NN_UNUSED nn_req_handle *hndl,
+    NN_UNUSED void *buf, NN_UNUSED size_t len, NN_UNUSED int flags)
+{
+    nn_assert (0);
+}
+
+
+int nn_req_crecv (struct nn_sockbase *self, struct nn_msg *msg)
 {
     struct nn_req *req;
 
@@ -317,8 +285,8 @@ static int nn_req_recv (struct nn_sockbase *self, struct nn_msg *msg)
         return -EAGAIN;
 
     /*  If the reply was already received, just pass it to the caller. */
-    nn_msg_mv (msg, &req->reply);
-    nn_msg_init (&req->reply, 0);
+    nn_msg_mv (msg, &req->task.reply);
+    nn_msg_init (&req->task.reply, 0);
 
     /*  Notify the state machine. */
     nn_fsm_action (&req->fsm, NN_REQ_ACTION_RECEIVED);
@@ -326,7 +294,7 @@ static int nn_req_recv (struct nn_sockbase *self, struct nn_msg *msg)
     return 0;
 }
 
-static int nn_req_setopt (struct nn_sockbase *self, int level, int option,
+int nn_req_setopt (struct nn_sockbase *self, int level, int option,
         const void *optval, size_t optvallen)
 {
     struct nn_req *req;
@@ -346,7 +314,7 @@ static int nn_req_setopt (struct nn_sockbase *self, int level, int option,
     return -ENOPROTOOPT;
 }
 
-static int nn_req_getopt (struct nn_sockbase *self, int level, int option,
+int nn_req_getopt (struct nn_sockbase *self, int level, int option,
         void *optval, size_t *optvallen)
 {
     struct nn_req *req;
@@ -367,7 +335,7 @@ static int nn_req_getopt (struct nn_sockbase *self, int level, int option,
     return -ENOPROTOOPT;
 }
 
-static void nn_req_shutdown (struct nn_fsm *self, int src, int type,
+void nn_req_shutdown (struct nn_fsm *self, int src, int type,
     NN_UNUSED void *srcptr)
 {
     struct nn_req *req;
@@ -375,11 +343,11 @@ static void nn_req_shutdown (struct nn_fsm *self, int src, int type,
     req = nn_cont (self, struct nn_req, fsm);
 
     if (nn_slow (src == NN_FSM_ACTION && type == NN_FSM_STOP)) {
-        nn_timer_stop (&req->timer);
+        nn_timer_stop (&req->task.timer);
         req->state = NN_REQ_STATE_STOPPING;
     }
     if (nn_slow (req->state == NN_REQ_STATE_STOPPING)) {
-        if (!nn_timer_isidle (&req->timer))
+        if (!nn_timer_isidle (&req->task.timer))
             return;
         req->state = NN_REQ_STATE_IDLE;
         nn_fsm_stopped_noevent (&req->fsm);
@@ -390,7 +358,7 @@ static void nn_req_shutdown (struct nn_fsm *self, int src, int type,
     nn_fsm_bad_state(req->state, src, type);
 }
 
-static void nn_req_handler (struct nn_fsm *self, int src, int type,
+void nn_req_handler (struct nn_fsm *self, int src, int type,
     NN_UNUSED void *srcptr)
 {
     struct nn_req *req;
@@ -476,8 +444,8 @@ static void nn_req_handler (struct nn_fsm *self, int src, int type,
             case NN_REQ_ACTION_IN:
 
                 /*  Reply arrived. */
-                nn_timer_stop (&req->timer);
-                req->sent_to = NULL;
+                nn_timer_stop (&req->task.timer);
+                req->task.sent_to = NULL;
                 req->state = NN_REQ_STATE_STOPPING_TIMER;
                 return;
 
@@ -485,15 +453,15 @@ static void nn_req_handler (struct nn_fsm *self, int src, int type,
 
                 /*  New request was sent while the old one was still being
                     processed. Cancel the old request first. */
-                nn_timer_stop (&req->timer);
-                req->sent_to = NULL;
+                nn_timer_stop (&req->task.timer);
+                req->task.sent_to = NULL;
                 req->state = NN_REQ_STATE_CANCELLING;
                 return;
 
             case NN_REQ_ACTION_PIPE_RM:
                 /*  Pipe that we sent request to is removed  */
-                nn_timer_stop (&req->timer);
-                req->sent_to = NULL;
+                nn_timer_stop (&req->task.timer);
+                req->task.sent_to = NULL;
                 /*  Pretend we timed out so request resent immediately  */
                 req->state = NN_REQ_STATE_TIMED_OUT;
                 return;
@@ -505,8 +473,8 @@ static void nn_req_handler (struct nn_fsm *self, int src, int type,
         case NN_REQ_SRC_RESEND_TIMER:
             switch (type) {
             case NN_TIMER_TIMEOUT:
-                nn_timer_stop (&req->timer);
-                req->sent_to = NULL;
+                nn_timer_stop (&req->task.timer);
+                req->task.sent_to = NULL;
                 req->state = NN_REQ_STATE_TIMED_OUT;
                 return;
             default:
@@ -650,14 +618,14 @@ static void nn_req_handler (struct nn_fsm *self, int src, int type,
 /*  State machine actions.                                                    */
 /******************************************************************************/
 
-static void nn_req_action_send (struct nn_req *self, int allow_delay)
+void nn_req_action_send (struct nn_req *self, int allow_delay)
 {
     int rc;
     struct nn_msg msg;
     struct nn_pipe *to;
 
     /*  Send the request. */
-    nn_msg_cp (&msg, &self->request);
+    nn_msg_cp (&msg, &self->task.request);
     rc = nn_xreq_send_to (&self->xreq.sockbase, &msg, &to);
 
     /*  If the request cannot be sent at the moment wait till
@@ -673,9 +641,9 @@ static void nn_req_action_send (struct nn_req *self, int allow_delay)
         in case the request gets lost somewhere further out
         in the topology. */
     if (nn_fast (rc == 0)) {
-        nn_timer_start (&self->timer, self->resend_ivl);
+        nn_timer_start (&self->task.timer, self->resend_ivl);
         nn_assert (to);
-        self->sent_to = to;
+        self->task.sent_to = to;
         self->state = NN_REQ_STATE_ACTIVE;
         return;
     }
@@ -702,7 +670,7 @@ void nn_req_rm (struct nn_sockbase *self, struct nn_pipe *pipe) {
     req = nn_cont (self, struct nn_req, xreq.sockbase);
 
     nn_xreq_rm (self, pipe);
-    if (nn_slow (pipe == req->sent_to)) {
+    if (nn_slow (pipe == req->task.sent_to)) {
         nn_fsm_action (&req->fsm, NN_REQ_ACTION_PIPE_RM);
     }
 }
