@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2012 250bpm s.r.o.  All rights reserved.
+    Copyright (c) 2012 Martin Sustrik  All rights reserved.
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"),
@@ -23,6 +23,8 @@
 #ifndef REQREP_H_INCLUDED
 #define REQREP_H_INCLUDED
 
+#include "nn.h"
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -34,6 +36,14 @@ extern "C" {
 
 #define NN_REQ_RESEND_IVL 1
 
+typedef union nn_req_handle {
+    int i;
+    void *ptr;
+} nn_req_handle;
+
+NN_EXPORT int nn_req_send (int s, nn_req_handle hndl, const void *buf, size_t len, int flags);
+NN_EXPORT int nn_req_recv (int s, nn_req_handle *hndl, void *buf, size_t len, int flags);
+
 #ifdef __cplusplus
 }
 #endif
